# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: CC-BY-NC-ND-4.0 OR LicenseRef-Proprietary
# Please refer to the accompanying LICENSE file.

# distutils: language=c

cdef extern from "<stdarg.h>":

    ctypedef struct va_list:
        pass

#cdef extern from *:
#cdef public:
#cdef public api: # from "jni.h":
cdef api: # from "jni.h":

    #
    # JNI Types
    #

    ctypedef long           c_jint
    ctypedef long long      c_jlong  # int64_t, jni.h: __int64
    ctypedef signed char    c_jbyte  # javabridge: unsigned (chyba blad!!!)

    ctypedef unsigned char  c_jboolean
    ctypedef unsigned short c_jchar
    ctypedef short          c_jshort
    ctypedef float          c_jfloat
    ctypedef double         c_jdouble
    ctypedef c_jint         c_jsize

    struct _jobject
    ctypedef _jobject* c_jobject
    ctypedef c_jobject c_jclass
    ctypedef c_jobject c_jthrowable
    ctypedef c_jobject c_jstring
    ctypedef c_jobject c_jarray
    ctypedef c_jarray  c_jbooleanArray
    ctypedef c_jarray  c_jbyteArray
    ctypedef c_jarray  c_jcharArray
    ctypedef c_jarray  c_jshortArray
    ctypedef c_jarray  c_jintArray
    ctypedef c_jarray  c_jlongArray
    ctypedef c_jarray  c_jfloatArray
    ctypedef c_jarray  c_jdoubleArray
    ctypedef c_jarray  c_jobjectArray
    ctypedef c_jobject c_jweak

    ctypedef union c_jvalue:
        c_jboolean z
        c_jbyte    b
        c_jchar    c
        c_jshort   s
        c_jint     i
        c_jlong    j
        c_jfloat   f
        c_jdouble  d
        c_jobject  l

    struct _jfieldID
    ctypedef _jfieldID* c_jfieldID

    struct _jmethodID
    ctypedef _jmethodID* c_jmethodID

    # Return values from jobjectRefType

    cpdef enum:
        JNIInvalidRefType    = 0
        JNILocalRefType      = 1
        JNIGlobalRefType     = 2
        JNIWeakGlobalRefType = 3
    ctypedef int c_jobjectRefType

    #
    # jboolean constants
    #

    cpdef enum:
        JNI_FALSE = 0
        JNI_TRUE  = 1

    #
    # possible return values for JNI functions.
    #

    cpdef enum:
        JNI_OK        =  0  # success
        JNI_ERR       = -1  # unknown error
        JNI_EDETACHED = -2  # thread detached from the VM
        JNI_EVERSION  = -3  # JNI version error
        JNI_ENOMEM    = -4  # not enough memory
        JNI_EEXIST    = -5  # VM already created
        JNI_EINVAL    = -6  # invalid arguments

    #
    # used in ReleaseScalarArrayElements
    #

    cpdef enum:
        JNI_COMMIT = 1
        JNI_ABORT  = 2

    #
    # used in RegisterNatives to describe native method name,
    # signature, and function pointer.
    #

    ctypedef struct c_JNINativeMethod:
        char* name       # Android and/or jnius: const char*
        char* signature  # Android and/or jnius: const char*
        void* fnPtr

    #
    # JNI Native Method Interface.
    #

    struct JNINativeInterface_
    ctypedef const JNINativeInterface_* c_JNIEnv

    #
    # JNI Invocation Interface.
    #

    struct JNIInvokeInterface_
    ctypedef const JNIInvokeInterface_* c_JavaVM

    # JNI Native Method Interface.

    struct JNINativeInterface_:

        void* reserved0
        void* reserved1
        void* reserved2
        void* reserved3

        c_jint        (*GetVersion)                   (c_JNIEnv* env)

        c_jclass      (*DefineClass)                  (c_JNIEnv* env, const char* name, c_jobject loader, const c_jbyte* buf, c_jsize len)
        c_jclass      (*FindClass)                    (c_JNIEnv* env, const char* name)

        c_jmethodID   (*FromReflectedMethod)          (c_JNIEnv* env, c_jobject method)
        c_jfieldID    (*FromReflectedField)           (c_JNIEnv* env, c_jobject field)

        c_jobject     (*ToReflectedMethod)            (c_JNIEnv* env, c_jclass cls, c_jmethodID methodID, c_jboolean isStatic)

        c_jclass      (*GetSuperclass)                (c_JNIEnv* env, c_jclass sub)
        c_jboolean    (*IsAssignableFrom)             (c_JNIEnv* env, c_jclass sub, c_jclass sup)

        c_jobject     (*ToReflectedField)             (c_JNIEnv* env, c_jclass cls, c_jfieldID fieldID, c_jboolean isStatic)

        c_jint        (*Throw)                        (c_JNIEnv* env, c_jthrowable obj)
        c_jint        (*ThrowNew)                     (c_JNIEnv* env, c_jclass clazz, const char* msg)
        c_jthrowable  (*ExceptionOccurred)            (c_JNIEnv* env)
        void          (*ExceptionDescribe)            (c_JNIEnv* env)
        void          (*ExceptionClear)               (c_JNIEnv* env)
        void          (*FatalError)                   (c_JNIEnv* env, const char* msg)

        c_jint        (*PushLocalFrame)               (c_JNIEnv* env, c_jint capacity)
        c_jobject     (*PopLocalFrame)                (c_JNIEnv* env, c_jobject result)

        c_jobject     (*NewGlobalRef)                 (c_JNIEnv* env, c_jobject lobj)
        void          (*DeleteGlobalRef)              (c_JNIEnv* env, c_jobject gref)
        void          (*DeleteLocalRef)               (c_JNIEnv* env, c_jobject obj)
        c_jboolean    (*IsSameObject)                 (c_JNIEnv* env, c_jobject obj1, c_jobject obj2)
        c_jobject     (*NewLocalRef)                  (c_JNIEnv* env, c_jobject ref)
        c_jint        (*EnsureLocalCapacity)          (c_JNIEnv* env, c_jint capacity)

        c_jobject     (*AllocObject)                  (c_JNIEnv* env, c_jclass clazz)
        c_jobject     (*NewObject)                    (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)
        c_jobject     (*NewObjectV)                   (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)
        c_jobject     (*NewObjectA)                   (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)

        c_jclass      (*GetObjectClass)               (c_JNIEnv* env, c_jobject obj)
        c_jboolean    (*IsInstanceOf)                 (c_JNIEnv* env, c_jobject obj, c_jclass clazz)

        c_jmethodID   (*GetMethodID)                  (c_JNIEnv* env, c_jclass clazz, const char* name, const char* sig)

        c_jobject     (*CallObjectMethod)             (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jobject     (*CallObjectMethodV)            (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jobject     (*CallObjectMethodA)            (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jboolean    (*CallBooleanMethod)            (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jboolean    (*CallBooleanMethodV)           (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jboolean    (*CallBooleanMethodA)           (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jbyte       (*CallByteMethod)               (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jbyte       (*CallByteMethodV)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jbyte       (*CallByteMethodA)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jchar       (*CallCharMethod)               (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jchar       (*CallCharMethodV)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jchar       (*CallCharMethodA)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jshort      (*CallShortMethod)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jshort      (*CallShortMethodV)             (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jshort      (*CallShortMethodA)             (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jint        (*CallIntMethod)                (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jint        (*CallIntMethodV)               (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jint        (*CallIntMethodA)               (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jlong       (*CallLongMethod)               (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jlong       (*CallLongMethodV)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jlong       (*CallLongMethodA)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jfloat      (*CallFloatMethod)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jfloat      (*CallFloatMethodV)             (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jfloat      (*CallFloatMethodA)             (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jdouble     (*CallDoubleMethod)             (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        c_jdouble     (*CallDoubleMethodV)            (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        c_jdouble     (*CallDoubleMethodA)            (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        void          (*CallVoidMethod)               (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, ...)  # nogil
        void          (*CallVoidMethodV)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, va_list args)  # nogil
        void          (*CallVoidMethodA)              (c_JNIEnv* env, c_jobject obj, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jobject     (*CallNonvirtualObjectMethod)   (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jobject     (*CallNonvirtualObjectMethodV)  (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jobject     (*CallNonvirtualObjectMethodA)  (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jboolean    (*CallNonvirtualBooleanMethod)  (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jboolean    (*CallNonvirtualBooleanMethodV) (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jboolean    (*CallNonvirtualBooleanMethodA) (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jbyte       (*CallNonvirtualByteMethod)     (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jbyte       (*CallNonvirtualByteMethodV)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jbyte       (*CallNonvirtualByteMethodA)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jchar       (*CallNonvirtualCharMethod)     (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jchar       (*CallNonvirtualCharMethodV)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jchar       (*CallNonvirtualCharMethodA)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jshort      (*CallNonvirtualShortMethod)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jshort      (*CallNonvirtualShortMethodV)   (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jshort      (*CallNonvirtualShortMethodA)   (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jint        (*CallNonvirtualIntMethod)      (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jint        (*CallNonvirtualIntMethodV)     (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jint        (*CallNonvirtualIntMethodA)     (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jlong       (*CallNonvirtualLongMethod)     (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jlong       (*CallNonvirtualLongMethodV)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jlong       (*CallNonvirtualLongMethodA)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jfloat      (*CallNonvirtualFloatMethod)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jfloat      (*CallNonvirtualFloatMethodV)   (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jfloat      (*CallNonvirtualFloatMethodA)   (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jdouble     (*CallNonvirtualDoubleMethod)   (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jdouble     (*CallNonvirtualDoubleMethodV)  (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jdouble     (*CallNonvirtualDoubleMethodA)  (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        void          (*CallNonvirtualVoidMethod)     (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        void          (*CallNonvirtualVoidMethodV)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        void          (*CallNonvirtualVoidMethodA)    (c_JNIEnv* env, c_jobject obj, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jfieldID    (*GetFieldID)                   (c_JNIEnv* env, c_jclass clazz, const char* name, const char* sig)

        c_jobject     (*GetObjectField)               (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jboolean    (*GetBooleanField)              (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jbyte       (*GetByteField)                 (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jchar       (*GetCharField)                 (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jshort      (*GetShortField)                (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jint        (*GetIntField)                  (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jlong       (*GetLongField)                 (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jfloat      (*GetFloatField)                (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)
        c_jdouble     (*GetDoubleField)               (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID)

        void          (*SetObjectField)               (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jobject  value)
        void          (*SetBooleanField)              (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jboolean value)
        void          (*SetByteField)                 (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jbyte    value)
        void          (*SetCharField)                 (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jchar    value)
        void          (*SetShortField)                (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jshort   value)
        void          (*SetIntField)                  (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jint     value)
        void          (*SetLongField)                 (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jlong    value)
        void          (*SetFloatField)                (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jfloat   value)
        void          (*SetDoubleField)               (c_JNIEnv* env, c_jobject obj, c_jfieldID fieldID, c_jdouble  value)

        c_jmethodID   (*GetStaticMethodID)            (c_JNIEnv* env, c_jclass clazz, const char* name, const char* sig)

        c_jobject     (*CallStaticObjectMethod)       (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jobject     (*CallStaticObjectMethodV)      (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jobject     (*CallStaticObjectMethodA)      (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jboolean    (*CallStaticBooleanMethod)      (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jboolean    (*CallStaticBooleanMethodV)     (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jboolean    (*CallStaticBooleanMethodA)     (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jbyte       (*CallStaticByteMethod)         (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jbyte       (*CallStaticByteMethodV)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jbyte       (*CallStaticByteMethodA)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jchar       (*CallStaticCharMethod)         (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jchar       (*CallStaticCharMethodV)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jchar       (*CallStaticCharMethodA)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jshort      (*CallStaticShortMethod)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jshort      (*CallStaticShortMethodV)       (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jshort      (*CallStaticShortMethodA)       (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jint        (*CallStaticIntMethod)          (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jint        (*CallStaticIntMethodV)         (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jint        (*CallStaticIntMethodA)         (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jlong       (*CallStaticLongMethod)         (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jlong       (*CallStaticLongMethodV)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jlong       (*CallStaticLongMethodA)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jfloat      (*CallStaticFloatMethod)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jfloat      (*CallStaticFloatMethodV)       (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jfloat      (*CallStaticFloatMethodA)       (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jdouble     (*CallStaticDoubleMethod)       (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        c_jdouble     (*CallStaticDoubleMethodV)      (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        c_jdouble     (*CallStaticDoubleMethodA)      (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        void          (*CallStaticVoidMethod)         (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, ...)  # nogil
        void          (*CallStaticVoidMethodV)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, va_list args)  # nogil
        void          (*CallStaticVoidMethodA)        (c_JNIEnv* env, c_jclass clazz, c_jmethodID methodID, const c_jvalue* args)  # nogil

        c_jfieldID    (*GetStaticFieldID)             (c_JNIEnv* env, c_jclass clazz, const char* name, const char* sig)

        c_jobject     (*GetStaticObjectField)         (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jboolean    (*GetStaticBooleanField)        (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jbyte       (*GetStaticByteField)           (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jchar       (*GetStaticCharField)           (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jshort      (*GetStaticShortField)          (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jint        (*GetStaticIntField)            (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jlong       (*GetStaticLongField)           (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jfloat      (*GetStaticFloatField)          (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)
        c_jdouble     (*GetStaticDoubleField)         (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID)

        void          (*SetStaticObjectField)         (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jobject  value)
        void          (*SetStaticBooleanField)        (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jboolean value)
        void          (*SetStaticByteField)           (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jbyte    value)
        void          (*SetStaticCharField)           (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jchar    value)
        void          (*SetStaticShortField)          (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jshort   value)
        void          (*SetStaticIntField)            (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jint     value)
        void          (*SetStaticLongField)           (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jlong    value)
        void          (*SetStaticFloatField)          (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jfloat   value)
        void          (*SetStaticDoubleField)         (c_JNIEnv* env, c_jclass clazz, c_jfieldID fieldID, c_jdouble  value)

        c_jstring     (*NewString)                    (c_JNIEnv* env, const c_jchar* ustr, c_jsize len)
        c_jsize       (*GetStringLength)              (c_JNIEnv* env, c_jstring str)
        const c_jchar*(*GetStringChars)               (c_JNIEnv* env, c_jstring str, c_jboolean* isCopy)
        void          (*ReleaseStringChars)           (c_JNIEnv* env, c_jstring str, const c_jchar* chars)

        c_jstring     (*NewStringUTF)                 (c_JNIEnv* env, const char* utf)
        c_jsize       (*GetStringUTFLength)           (c_JNIEnv* env, c_jstring str)
        const char*   (*GetStringUTFChars)            (c_JNIEnv* env, c_jstring str, c_jboolean* isCopy)
        void          (*ReleaseStringUTFChars)        (c_JNIEnv* env, c_jstring str, const char* chars)

        c_jsize       (*GetArrayLength)               (c_JNIEnv* env, c_jarray array)

        c_jobjectArray (*NewObjectArray)              (c_JNIEnv* env, c_jsize len, c_jclass clazz, c_jobject init)
        c_jobject     (*GetObjectArrayElement)        (c_JNIEnv* env, c_jobjectArray array, c_jsize index)
        void          (*SetObjectArrayElement)        (c_JNIEnv* env, c_jobjectArray array, c_jsize index, c_jobject value)

        c_jbooleanArray (*NewBooleanArray)            (c_JNIEnv* env, c_jsize len)
        c_jbyteArray  (*NewByteArray)                 (c_JNIEnv* env, c_jsize len)
        c_jcharArray  (*NewCharArray)                 (c_JNIEnv* env, c_jsize len)
        c_jshortArray (*NewShortArray)                (c_JNIEnv* env, c_jsize len)
        c_jintArray   (*NewIntArray)                  (c_JNIEnv* env, c_jsize len)
        c_jlongArray  (*NewLongArray)                 (c_JNIEnv* env, c_jsize len)
        c_jfloatArray (*NewFloatArray)                (c_JNIEnv* env, c_jsize len)
        c_jdoubleArray(*NewDoubleArray)               (c_JNIEnv* env, c_jsize len)

        c_jboolean*   (*GetBooleanArrayElements)      (c_JNIEnv* env, c_jbooleanArray array, c_jboolean* isCopy)
        c_jbyte*      (*GetByteArrayElements)         (c_JNIEnv* env, c_jbyteArray    array, c_jboolean* isCopy)
        c_jchar*      (*GetCharArrayElements)         (c_JNIEnv* env, c_jcharArray    array, c_jboolean* isCopy)
        c_jshort*     (*GetShortArrayElements)        (c_JNIEnv* env, c_jshortArray   array, c_jboolean* isCopy)
        c_jint*       (*GetIntArrayElements)          (c_JNIEnv* env, c_jintArray     array, c_jboolean* isCopy)
        c_jlong*      (*GetLongArrayElements)         (c_JNIEnv* env, c_jlongArray    array, c_jboolean* isCopy)
        c_jfloat*     (*GetFloatArrayElements)        (c_JNIEnv* env, c_jfloatArray   array, c_jboolean* isCopy)
        c_jdouble*    (*GetDoubleArrayElements)       (c_JNIEnv* env, c_jdoubleArray  array, c_jboolean* isCopy)

        void          (*ReleaseBooleanArrayElements)  (c_JNIEnv* env, c_jbooleanArray array, c_jboolean* elems, c_jint mode)
        void          (*ReleaseByteArrayElements)     (c_JNIEnv* env, c_jbyteArray    array, c_jbyte*    elems, c_jint mode)
        void          (*ReleaseCharArrayElements)     (c_JNIEnv* env, c_jcharArray    array, c_jchar*    elems, c_jint mode)
        void          (*ReleaseShortArrayElements)    (c_JNIEnv* env, c_jshortArray   array, c_jshort*   elems, c_jint mode)
        void          (*ReleaseIntArrayElements)      (c_JNIEnv* env, c_jintArray     array, c_jint*     elems, c_jint mode)
        void          (*ReleaseLongArrayElements)     (c_JNIEnv* env, c_jlongArray    array, c_jlong*    elems, c_jint mode)
        void          (*ReleaseFloatArrayElements)    (c_JNIEnv* env, c_jfloatArray   array, c_jfloat*   elems, c_jint mode)
        void          (*ReleaseDoubleArrayElements)   (c_JNIEnv* env, c_jdoubleArray  array, c_jdouble*  elems, c_jint mode)

        void          (*GetBooleanArrayRegion)        (c_JNIEnv* env, c_jbooleanArray array, c_jsize start, c_jsize len, c_jboolean* buf)
        void          (*GetByteArrayRegion)           (c_JNIEnv* env, c_jbyteArray    array, c_jsize start, c_jsize len, c_jbyte*    buf)
        void          (*GetCharArrayRegion)           (c_JNIEnv* env, c_jcharArray    array, c_jsize start, c_jsize len, c_jchar*    buf)
        void          (*GetShortArrayRegion)          (c_JNIEnv* env, c_jshortArray   array, c_jsize start, c_jsize len, c_jshort*   buf)
        void          (*GetIntArrayRegion)            (c_JNIEnv* env, c_jintArray     array, c_jsize start, c_jsize len, c_jint*     buf)
        void          (*GetLongArrayRegion)           (c_JNIEnv* env, c_jlongArray    array, c_jsize start, c_jsize len, c_jlong*    buf)
        void          (*GetFloatArrayRegion)          (c_JNIEnv* env, c_jfloatArray   array, c_jsize start, c_jsize len, c_jfloat*   buf)
        void          (*GetDoubleArrayRegion)         (c_JNIEnv* env, c_jdoubleArray  array, c_jsize start, c_jsize len, c_jdouble*  buf)

        void          (*SetBooleanArrayRegion)        (c_JNIEnv* env, c_jbooleanArray array, c_jsize start, c_jsize len, const c_jboolean* buf)
        void          (*SetByteArrayRegion)           (c_JNIEnv* env, c_jbyteArray    array, c_jsize start, c_jsize len, const c_jbyte*    buf)
        void          (*SetCharArrayRegion)           (c_JNIEnv* env, c_jcharArray    array, c_jsize start, c_jsize len, const c_jchar*    buf)
        void          (*SetShortArrayRegion)          (c_JNIEnv* env, c_jshortArray   array, c_jsize start, c_jsize len, const c_jshort*   buf)
        void          (*SetIntArrayRegion)            (c_JNIEnv* env, c_jintArray     array, c_jsize start, c_jsize len, const c_jint*     buf)
        void          (*SetLongArrayRegion)           (c_JNIEnv* env, c_jlongArray    array, c_jsize start, c_jsize len, const c_jlong*    buf)
        void          (*SetFloatArrayRegion)          (c_JNIEnv* env, c_jfloatArray   array, c_jsize start, c_jsize len, const c_jfloat*   buf)
        void          (*SetDoubleArrayRegion)         (c_JNIEnv* env, c_jdoubleArray  array, c_jsize start, c_jsize len, const c_jdouble*  buf)

        c_jint        (*RegisterNatives)              (c_JNIEnv* env, c_jclass clazz, const c_JNINativeMethod* methods, c_jint nMethods)
        c_jint        (*UnregisterNatives)            (c_JNIEnv* env, c_jclass clazz)

        c_jint        (*MonitorEnter)                 (c_JNIEnv* env, c_jobject obj)
        c_jint        (*MonitorExit)                  (c_JNIEnv* env, c_jobject obj)

        c_jint        (*GetJavaVM)                    (c_JNIEnv* env, c_JavaVM** vm)

        void          (*GetStringRegion)              (c_JNIEnv* env, c_jstring str, c_jsize start, c_jsize len, c_jchar* buf)
        void          (*GetStringUTFRegion)           (c_JNIEnv* env, c_jstring str, c_jsize start, c_jsize len, char*    buf)

        void*         (*GetPrimitiveArrayCritical)    (c_JNIEnv* env, c_jarray array, c_jboolean* isCopy)
        void          (*ReleasePrimitiveArrayCritical)(c_JNIEnv* env, c_jarray array, void* carray, c_jint mode)

        const c_jchar*(*GetStringCritical)            (c_JNIEnv* env, c_jstring string, c_jboolean* isCopy)
        void          (*ReleaseStringCritical)        (c_JNIEnv* env, c_jstring string, const c_jchar* cstring)

        c_jweak       (*NewWeakGlobalRef)             (c_JNIEnv* env, c_jobject obj)
        void          (*DeleteWeakGlobalRef)          (c_JNIEnv* env, c_jweak   ref)

        c_jboolean    (*ExceptionCheck)               (c_JNIEnv* env)

        c_jobject     (*NewDirectByteBuffer)          (c_JNIEnv* env, void* address, c_jlong capacity)
        void*         (*GetDirectBufferAddress)       (c_JNIEnv* env, c_jobject buf)
        c_jlong       (*GetDirectBufferCapacity)      (c_JNIEnv* env, c_jobject buf)

        # New JNI 1.6 Features

        c_jobjectRefType (*GetObjectRefType)          (c_JNIEnv* env, c_jobject obj)

    # JNI Invocation Interface.

    ctypedef struct c_JavaVMOption:
        char* optionString
        void* extraInfo

    ctypedef struct c_JavaVMInitArgs:
        c_jint          version
        c_jint          nOptions
        c_JavaVMOption* options
        c_jboolean      ignoreUnrecognized

    ctypedef struct c_JavaVMAttachArgs:
        c_jint    version
        char*     name
        c_jobject group

    # These will be VM-specific.

    cpdef enum:
        JDK1_2 = 1
        JDK1_4 = 1

    # End VM-specific.

    struct JNIInvokeInterface_:

        void* reserved0
        void* reserved1
        void* reserved2

        c_jint (*DestroyJavaVM)              (c_JavaVM* vm)  # nogil
        c_jint (*AttachCurrentThread)        (c_JavaVM* vm, void** penv, void* args)  # nogil
        c_jint (*DetachCurrentThread)        (c_JavaVM* vm)  # nogil
        c_jint (*GetEnv)                     (c_JavaVM* vm, void** penv, c_jint version)  # nogil
        c_jint (*AttachCurrentThreadAsDaemon)(c_JavaVM* vm, void** penv, void* args)  # nogil

    cpdef enum:
        JNI_VERSION_1_1 = 0x00010001
        JNI_VERSION_1_2 = 0x00010002
        JNI_VERSION_1_4 = 0x00010004
        JNI_VERSION_1_6 = 0x00010006
        JNI_VERSION_1_8 = 0x00010008
        JNI_VERSION_9   = 0x00090000
        JNI_VERSION_10  = 0x000a0000

# eof jni.h

cdef extern: # from "jni.h":
    c_jint __stdcall JNI_GetDefaultJavaVMInitArgs(void* args)  # nogil
    c_jint __stdcall JNI_CreateJavaVM     (c_JavaVM** pvm, void** penv, void* args)  # nogil
    c_jint __stdcall JNI_GetCreatedJavaVMs(c_JavaVM** pvm, c_jsize size, c_jsize* nvms)  # nogil



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: CC-BY-NC-ND-4.0 OR LicenseRef-Proprietary
# Please refer to the accompanying LICENSE file.

from __future__ import annotations

import unittest
import re


class UtilTestCaseCFFI(unittest.TestCase):

    def test_preprocessor(self):

        from jni._util import Preprocessor

        char_pattern = Preprocessor.basic_tokens["char"]
        char_regexp  = re.compile(r"^" + char_pattern + r"$")

        vals = [
            r"'a'",
            r"L'a'",
            r"'$'",
            r"L'$'",
            r"'*'",
            r"L'*'",
            r"'\''",
            r"'\"'",
            r"'\\'",
            r"'\a'",
            r"'\b'",
            r"'\f'",
            r"'\n'",
            r"'\r'",
            r"'\t'",
            r"'\v'",
            r"'\0'",
            r"'\06'",
            r"'\075'",
            r"'\x0'",
            r"'\x0aF'",
            r"'\x0aF23B87c98'",
        ]

        for val in vals:
            self.assertTrue(char_regexp.match(val))



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: CC-BY-NC-ND-4.0 OR LicenseRef-Proprietary
# Please refer to the accompanying LICENSE file.

from typing import Optional, Tuple
from pathlib import Path
import os
import enum

import jni

class EStatusCode(enum.IntEnum):
    SUCCESS   = 1000
    UNKNOWN   = 1001
    EXCEPTION = 1002
    HOST      = 1003
    ERR       = jni.JNI_ERR        # unknown error
    EDETACHED = jni.JNI_EDETACHED  # thread detached from the VM
    EVERSION  = jni.JNI_EVERSION   # JNI version error
    ENOMEM    = jni.JNI_ENOMEM     # not enough memory
    EEXIST    = jni.JNI_EEXIST     # VM already created
    EINVAL    = jni.JNI_EINVAL     # invalid arguments


class JVM:

    """Java Virtual Machine"""

    JNI_VERSION = jni.JNI_VERSION_1_6

    def createdJVMs(self) -> Tuple['_JVM', ...]:
        njvm = jni.new(jni.jsize)
        err = self._jvm.JNI.GetCreatedJavaVMs(None, 0, njvm)
        if err != jni.JNI_OK:
            raise jni.JNIException(err, info="JNI_GetCreatedJavaVMs")
        pjvm = jni.new_array(jni.POINTER(jni.JavaVM), njvm[0])
        err = self._jvm.JNI.GetCreatedJavaVMs(pjvm, len(pjvm), njvm)
        if err != jni.JNI_OK:
            raise jni.JNIException(err, info="JNI_GetCreatedJavaVMs")
        jvms = []
        for i in range(njvm[0]):
            jvm = _JVM()
            jvm.JNI    = self._jvm.JNI
            jvm.jnijvm = pjvm[0][i]
            jvms.append(jvm)
        return tuple(jvms)

    def __init__(self, dll_path: Optional[object] = None):

        if_load = dll_path is not None

        self._jvm = None  # _JVM
        self.JavaException = None
        self.ExceptionsMap = {}
        try:
            # TODO: add the non-string parameters, for possible callbacks

            if if_load and not isinstance(dll_path, (str, os.PathLike)):
                raise JVMException(EStatusCode.EINVAL,
                                   "First parameter must be a string or os.PathLike type")
            self._jvm = _JVM()
            try:
                self._jvm.JNI = jni.load(dll_path) if if_load else None
            except Exception as exc:
                raise JVMException(EStatusCode.UNKNOWN,
                                   f"Unable to load DLL [{dll_path}], error = {exc}") from None
            self._jvm._create()
        except Exception as exc:
            self.handleException(exc)

    def __del__(self):
        if not self._jvm: return
        try: self._jvm.jnijvm.DestroyJavaVM()
        except Exception: pass
        self._jvm.jnijvm = None
        try: self._jvm.JNI.dllclose()
        except Exception: pass
        self._jvm.JNI = None

    def __enter__(self):
        if self._jvm is None:
            raise JVMException(EStatusCode.EDETACHED,
                               "Unable to use JVM: thread detached from the VM")
        if self._jvm.jnijvm:
            penv = jni.obj(jni.POINTER(jni.JNIEnv))
            self._jvm.jnijvm.AttachCurrentThread(penv)
            return self._jvm, jni.JEnv(penv)
        else:
            return self._jvm, None

    def __exit__(self, exc_type, exc, exc_tb):
        del exc_type, exc_tb
        if exc: self.handleException(exc)
        return True

    def __iter__(self):
        return iter(self.__enter__())

    def start(self, *jvmoptions, **jvmargs): # -> Tuple['_JVM', jni.JNIEnv]:
        ignoreUnrecognized = jvmargs.get("ignoreUnrecognized", True)
        try:
            pjvm = jni.obj(jni.POINTER(jni.JavaVM))
            penv = jni.obj(jni.POINTER(jni.JNIEnv))
            jvm_args = jni.obj(jni.JavaVMInitArgs)
            jvm_args.version  = JVM.JNI_VERSION
            jvm_args.nOptions = len(jvmoptions)
            jvm_args.options  = joptions = jni.new_array(jni.JavaVMOption, jvm_args.nOptions)
            _keep = []
            for i, option in enumerate(jvmoptions):
                optionString = jni.new_cstr(option if isinstance(option, bytes)
                                            else str(option).encode("utf-8"))
                _keep.append(optionString)
                jvm_args.options[i].optionString = optionString
                jvm_args.options[i].extraInfo    = jni.NULL
            jvm_args.ignoreUnrecognized = jni.JNI_TRUE if ignoreUnrecognized else jni.JNI_FALSE
            err = self._jvm.JNI.CreateJavaVM(pjvm, penv, jvm_args)
            del _keep, joptions, jvm_args
            if err != jni.JNI_OK or jni.isNULL(pjvm):
                raise jni.JNIException(err if err != jni.JNI_OK else jni.JNI_ERR,
                                       info="JNI_CreateJavaVM")
            self._jvm.jnijvm = jni.JVM(pjvm)
            jenv = jni.JEnv(penv)
            try:
                self._jvm._initialize(jenv)
            except Exception as exc:
                try: self._jvm.jnijvm.DestroyJavaVM()
                except Exception: pass
                raise exc
            return self._jvm, jenv
        except Exception as exc:
            try:
                self.handleException(exc)
            finally:
                self._jvm.jnijvm = None

    def attach(self, pjvm: Optional[object] = None): # -> Tuple['_JVM', jni.JNIEnv]:
        if_bind = pjvm is not None
        try:
            if if_bind and not pjvm:
                raise JVMException(EStatusCode.EINVAL,
                                   "First parameter must be a JNI jvm handle")
            penv = jni.obj(jni.POINTER(jni.JNIEnv))
            if if_bind:
                self._jvm.jnijvm = jni.cast(pjvm, jni.POINTER(jni.JavaVM))[0]
                self._jvm.jnijvm.AttachCurrentThread(penv)
            else:
                self._jvm.jnijvm.GetEnv(penv, JVM.JNI_VERSION)
            jenv = jni.JEnv(penv)
            self._jvm._initialize(jenv)
            return self._jvm, jenv
        except Exception as exc:
            try:
                self.handleException(exc)
            finally:
                if if_bind:
                    self._jvm.jnijvm = None

    def shutdown(self):
        if self._jvm.jnijvm is None: return
        try:
            penv = jni.obj(jni.POINTER(jni.JNIEnv))
            self._jvm.jnijvm.AttachCurrentThread(penv)
            jenv = jni.JEnv(penv)
            self._jvm._dispose(jenv)
            self._jvm.jnijvm.DestroyJavaVM()
        except Exception as exc:
            try:
                self.handleException(exc)
            finally:
                self._jvm.jnijvm = None

    def isStarted(self) -> bool:
        # Check if the JVM environment has been initialized
        return self._jvm is not None and self._jvm.jnijvm is not None

    def attachThread(self, daemon: bool=False):
        try:
            penv = jni.obj(jni.POINTER(jni.JNIEnv))
            if not daemon:
                self._jvm.jnijvm.AttachCurrentThread(penv)
            else:
                self._jvm.jnijvm.AttachCurrentThreadAsDaemon(penv)
            return self._jvm, jni.JEnv(penv)
        except Exception as exc:
            self.handleException(exc)

    def detachThread(self):
        try:
            self._jvm.jnijvm.DetachCurrentThread()
        except Exception as exc:
            self.handleException(exc)

    def isThreadAttached(self) -> bool:
        try:
            penv = jni.obj(jni.POINTER(jni.JNIEnv))
            self._jvm.jnijvm.GetEnv(penv, JVM.JNI_VERSION)
        except jni.JNIException as exc:
            if exc.getError() == jni.JNI_EDETACHED:
                return False
            self.handleException(exc)
        except Exception as exc:
            self.handleException(exc)
        else:
            return not jni.isNULL(penv)

    def handleException(self, exc):
        try:
            raise exc
        except jni.Throwable as exc:
            self.JException.printDescribe()
            jexc = self.JException(exc)
            if self.JavaException and hasattr(self.JavaException, "__exception__"):
                raise self.JavaException.__exception__(jexc) from None
            else:
                classname = jexc.getClass().getName()
                message   = jexc.getMessage()
                if message is None: message = classname
                PyExc = self.JavaException or RuntimeError
                raise PyExc(f"Java exception {classname} occurred: {message}") from None
        except jni.JNIException as exc:
            PyExc = self.ExceptionsMap.get(exc.getError(),
                                           self.ExceptionsMap.get(EStatusCode.ERR, RuntimeError))
            raise PyExc(exc.getMessage()) from None
        except JVMException as exc:
            PyExc = self.ExceptionsMap.get(exc.args[0],
                                           self.ExceptionsMap.get(EStatusCode.ERR, RuntimeError))
            raise PyExc(exc.args[1]) from None
        except Exception:
            PyExc = self.ExceptionsMap.get(None)
            raise (PyExc(exc) if PyExc else exc) from None


class JVMException(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class _JVM:

    def __init__(self):
        super().__init__()
        self.JNI    = None  # jni.JNI
        self.jnijvm = None  # jni.JavaVM

    def _create(self):
        pass

    def _initialize(self, jenv: jni.JNIEnv):
        pass

    def _dispose(self, jenv: jni.JNIEnv):
        pass



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: CC-BY-NC-ND-4.0 OR LicenseRef-Proprietary
# Please refer to the accompanying LICENSE file.

if __name__.rpartition(".")[-1] == "__main__":
    import sys
    from . __main import main
    # logging.basicConfig(level=logging.INFO)
    # logging.basicConfig(level=logging.DEBUG)
    sys.exit(main())



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: CC-BY-NC-ND-4.0 OR LicenseRef-Proprietary
# Please refer to the accompanying LICENSE file.

import sys
from os import path
from io import open
from glob import glob
from setuptools import setup
from setuptools import Extension

top_dir = path.dirname(path.abspath(__file__))

GEN_CFFI = not path.exists("src/jni/cffi/jni.py")
if GEN_CFFI:
    def build_ffi(src_dir, emit_jni_i=False):
        import sys
        from os import path
        import cffi
        sys.path.insert(0, src_dir)
        from jni._util import Preprocessor
        sys.path.pop(0)
        with open(path.join(src_dir, "jni/cffi/jni.h"),
                  "r", encoding="utf-8") as f:
            source = f.read()
        source = Preprocessor().preprocess(source,
            define_macros = (
                ("JNICALL",   ""),
                ("JNIIMPORT", ""),
                ("JNIEXPORT", ""),
            ),
            undef_macros = (
                "_JAVASOFT_JNI_H_",
                "__cplusplus",
                "JNI_TYPES_ALREADY_DEFINED_IN_JNI_MD_H",
                "_JNI_IMPLEMENTATION_",
            ))
        if emit_jni_i:
            with open(path.join(src_dir, "jni/cffi/jni.i"),
                      "w", encoding="utf-8") as f:
                f.write(source)
        ffi = cffi.FFI()
        ffi.set_source("jni", None)
        ffi.cdef("// temporary for cffi")
        ffi.cdef("typedef void* va_list;")
        ffi.cdef(source)
        ffi.compile(tmpdir=path.join(src_dir, "jni/cffi"))
    build_ffi(path.join(top_dir, "src"))

GEN_CYTHON = (not path.exists("src/jni/cython/jni.c") or
              not path.exists("src/jni/cython/jni.h"))
ext = ".pyx" if GEN_CYTHON else ".c"
ext_modules = [Extension(name="jni.cython.jni", language="c",
                         sources=["src/jni/cython/jni" + ext])]
if GEN_CYTHON:
    from Cython.Build    import cythonize
    from Cython.Compiler import Options
    Options.docstrings         = False
    Options.emit_code_comments = False
    ext_modules = cythonize(ext_modules,
                            language_level=sys.version_info.major,
                            force=True)#, show_all_warnings=True)

platform = ("linux" if sys.platform.startswith("linux") else
            "android" if hasattr(sys, "getandroidapilevel") else sys.platform)
ext_modules += [Extension(name="jni.capi.jni", language="c",
                          include_dirs=["src/jni/capi/java/jdk/include/" + platform],
                          sources=["src/jni/capi/jni.c"])]

setup(ext_modules=ext_modules)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: CC-BY-NC-ND-4.0 OR LicenseRef-Proprietary
# Please refer to the accompanying LICENSE file.

import sys
import os
import platform
import ctypes as ct

from .jni import ffi ; del jni  # noqa

DLL     = lambda name, handle=None, __ffi=ffi: __ffi.dlopen(name)
dlclose = lambda handle,            __ffi=ffi: __ffi.dlclose(handle)

def CFUNC(restype, *argtypes, __ffi=ffi):
    callback_sign = ((restype.cname if restype is not None else "void") + " __stdcall"
                     + " (" + ", ".join(arg_type.cname for arg_type in argtypes) + ")")
    return __ffi.callback(callback_sign)


tmap = {}
__none = object()

POINTER   = lambda ctype, __tmap=tmap,   __ffi=ffi: __ffi.typeof(__ffi.getctype(__tmap.get(ctype, ctype))+"*")
pointer   = lambda obj,                  __ffi=ffi: __ffi.new(POINTER(__ffi.typeof(obj)), obj)
byref     = lambda obj, offset=0,        __ffi=ffi: __ffi.new(POINTER(__ffi.typeof(obj)), obj)  # !!! zamplementowac offset !!!
addressof = lambda obj,                  __ffi=ffi: __ffi.addressof(obj)
cast      = lambda obj, ctype,           __ffi=ffi: __ffi.cast(ctype, obj)
sizeof    = lambda obj_or_ctype,         __ffi=ffi: __ffi.sizeof(obj_or_ctype)
py_object = lambda obj,                  __ffi=ffi: __ffi.new_handle(obj)
memmove   = lambda dst, src, count,      __ffi=ffi: __ffi.memmove(dst, src, count)
obj       = lambda ctype, init=__none,   __ffi=ffi: (__ffi.new(__ffi.getctype(ctype)+"*", None if init is __none else init)[0]
                                                    if ctype.kind != "primitive" else
                                                    __ffi.cast(ctype, 0 if init is __none else init))
new         = lambda ctype, init=__none, __ffi=ffi: __ffi.new(__ffi.getctype(ctype)+"*", None if init is __none else init)
new_array   = lambda ctype, size,        __ffi=ffi: __ffi.new(__ffi.getctype(ctype)+"[]", size)
new_cstr    = lambda init,               __ffi=ffi: __ffi.new("char[]", init)
as_cstr     = lambda obj,                __ffi=ffi: __ffi.cast("char*", __ffi.from_buffer(obj))
to_bytes    = lambda obj, size=-1,       __ffi=ffi: __ffi.string(obj, size) if size >= 0 else __ffi.string(obj)
to_unicode  = to_bytes
from_buffer = lambda data,               __ffi=ffi: __ffi.from_buffer(data)

_itself_or_NULL = lambda arg, __NULL=ffi.NULL: __NULL if arg is None else arg         # noqa: N816
_byref_or_NULL  = lambda arg, __NULL=ffi.NULL: __NULL if arg is None else byref(arg)  # noqa: N816

class _CData(ct.Structure):
    _fields_ = (  # PyObject_HEAD
    ("ob_refcnt",     ct.c_ssize_t),
    ("ob_type",       ct.c_void_p),
    # _CData body
    ("c_type",        ct.py_object),  # CTypeDescrObject*
    ("c_data",        ct.c_void_p),   # char*
    ("c_weakreflist", ct.py_object),  # PyObject*
)


_as_CData = lambda obj, __ct=ct: __ct.cast(id(obj), __ct.POINTER(_CData))[0]  # noqa: N816

def defined(varname, __getframe=sys._getframe):
    frame = __getframe(1)
    return varname in frame.f_locals or varname in frame.f_globals

def from_oid(oid, __cast=ct.cast, __py_object=ct.py_object):
    return __cast(oid, __py_object).value if oid else None


#
# JNI Types
#

jint     = ffi.typeof("jint")
jlong    = ffi.typeof("jlong")
jbyte    = ffi.typeof("jbyte")

jboolean = ffi.typeof("jboolean")
jchar    = ffi.typeof("jchar")
jshort   = ffi.typeof("jshort")
jfloat   = ffi.typeof("jfloat")
jdouble  = ffi.typeof("jdouble")
jsize    = ffi.typeof("jsize")

jobject       = ffi.typeof("jobject")
jclass        = ffi.typeof("jclass")
jthrowable    = ffi.typeof("jthrowable")
jstring       = ffi.typeof("jstring")
jarray        = ffi.typeof("jarray")
jbooleanArray = ffi.typeof("jbooleanArray")  # noqa: N816
jbyteArray    = ffi.typeof("jbyteArray")     # noqa: N816
jcharArray    = ffi.typeof("jcharArray")     # noqa: N816
jshortArray   = ffi.typeof("jshortArray")    # noqa: N816
jintArray     = ffi.typeof("jintArray")      # noqa: N816
jlongArray    = ffi.typeof("jlongArray")     # noqa: N816
jfloatArray   = ffi.typeof("jfloatArray")    # noqa: N816
jdoubleArray  = ffi.typeof("jdoubleArray")   # noqa: N816
jobjectArray  = ffi.typeof("jobjectArray")   # noqa: N816
jweak         = ffi.typeof("jweak")

jvalue = ffi.typeof("jvalue")

jfieldID = ffi.typeof("jfieldID")  # noqa: N816

jmethodID = ffi.typeof("jmethodID")  # noqa: N816

# Return values from jobjectRefType

jobjectRefType = ffi.typeof("jobjectRefType")  # noqa: N816
JNIInvalidRefType    = jobjectRefType.relements["JNIInvalidRefType"]
JNILocalRefType      = jobjectRefType.relements["JNILocalRefType"]
JNIGlobalRefType     = jobjectRefType.relements["JNIGlobalRefType"]
JNIWeakGlobalRefType = jobjectRefType.relements["JNIWeakGlobalRefType"]

#
# jboolean constants
#

JNI_FALSE = ffi.integer_const("JNI_FALSE")
JNI_TRUE  = ffi.integer_const("JNI_TRUE")

#
# null constant
#

NULL   = ffi.NULL
isNULL = lambda jobj, __NULL=NULL: jobj == __NULL  # noqa: N816

#
# possible return values for JNI functions.
#

JNI_OK        = ffi.integer_const("JNI_OK")         # success
JNI_ERR       = ffi.integer_const("JNI_ERR")        # unknown error
JNI_EDETACHED = ffi.integer_const("JNI_EDETACHED")  # thread detached from the VM
JNI_EVERSION  = ffi.integer_const("JNI_EVERSION")   # JNI version error
JNI_ENOMEM    = ffi.integer_const("JNI_ENOMEM")     # not enough memory
JNI_EEXIST    = ffi.integer_const("JNI_EEXIST")     # VM already created
JNI_EINVAL    = ffi.integer_const("JNI_EINVAL")     # invalid arguments

#
# used in ReleaseScalarArrayElements
#

JNI_COMMIT = ffi.integer_const("JNI_COMMIT")
JNI_ABORT  = ffi.integer_const("JNI_ABORT")

#
# JNI Native Method Interface.
#

# used in RegisterNatives to describe native method name,
# signature, and function pointer.

JNINativeMethod = ffi.typeof("JNINativeMethod")

# We use inlined functions for C++ so that programmers can write:
#
#    env->FindClass("java/lang/String")
#
# in C++ rather than:
#
#    (*env)->FindClass(env, "java/lang/String")
#
# in C.

class JNIEnv:

    def _handle_JNIException(self, err):
        import sys
        fun_name = sys._getframe(1).f_code.co_name
        raise JNIException(err, info=fun_name)

    def _handle_JavaException(self):
        env = self.__env
        fun = env[0]
        jthr = fun.ExceptionOccurred(env)
        fun.ExceptionClear(env)
        jexc = fun.NewGlobalRef(env, jthr)
        fun.DeleteLocalRef(env, jthr)
        if Throwable.last:
            cause = Throwable.last.getCause()
            if cause: fun.DeleteGlobalRef(env, cause)
            Throwable.last = None
        fun.ExceptionClear(env)
        Throwable.last = thr = Throwable(jexc)  # !!!, fname)
        raise thr

    # Java version

    def GetVersion(self):
        env = self.__env
        fun = env[0]
        ret = fun.GetVersion(env)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Java class handling

    def DefineClass(self, name, loader, buf, blen):
        env = self.__env
        fun = env[0]
        ret = fun.DefineClass(env, name, loader, cast(buf, "const jbyte*"), blen)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def FindClass(self, name):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.FindClass(env, name)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetSuperclass(self, sub):
        env = self.__env
        fun = env[0]
        ret = fun.GetSuperclass(env, sub)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret if not isNULL(ret) else None

    def IsAssignableFrom(self, sub, sup):
        env = self.__env
        fun = env[0]
        return bool(fun.IsAssignableFrom(env, sub, sup))

    # Java exceptions handling

    def Throw(self, obj):
        env = self.__env
        fun = env[0]
        ret = fun.Throw(env, obj)
        if ret != 0: self._handle_JNIException(ret)

    def ThrowNew(self, clazz, msg):
        env = self.__env
        fun = env[0]
        ret = fun.ThrowNew(env, clazz, msg)
        if ret != 0: self._handle_JNIException(ret)

    def ExceptionOccurred(self):
        env = self.__env
        fun = env[0]
        return fun.ExceptionOccurred(env)

    def ExceptionDescribe(self):
        env = self.__env
        fun = env[0]
        fun.ExceptionDescribe(env)

    def ExceptionClear(self):
        env = self.__env
        fun = env[0]
        if Throwable.last:
            cause = Throwable.last.getCause()
            if cause: fun.DeleteGlobalRef(env, cause)
            Throwable.last = None
        fun.ExceptionClear(env)

    def FatalError(self, msg):
        env = self.__env
        fun = env[0]
        fun.FatalError(env, msg)

    def ExceptionCheck(self):
        env = self.__env
        fun = env[0]
        return bool(fun.ExceptionCheck(env))

    # JVM Call frame

    def PushLocalFrame(self, capacity):
        env = self.__env
        fun = env[0]
        ret = fun.PushLocalFrame(env, capacity)
        if ret != 0 and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def PopLocalFrame(self, result):
        env = self.__env
        fun = env[0]
        return fun.PopLocalFrame(env, result)

    # Java references handling

    def NewGlobalRef(self, lobj):
        env = self.__env
        fun = env[0]
        ret = fun.NewGlobalRef(env, lobj)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def DeleteGlobalRef(self, gref):
        env = self.__env
        fun = env[0]
        if gref: fun.DeleteGlobalRef(env, gref)

    def NewLocalRef(self, ref):
        env = self.__env
        fun = env[0]
        ret = fun.NewLocalRef(env, ref)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def DeleteLocalRef(self, obj):
        env = self.__env
        fun = env[0]
        if obj: fun.DeleteLocalRef(env, obj)

    def NewWeakGlobalRef(self, obj):
        env = self.__env
        fun = env[0]
        ret = fun.NewWeakGlobalRef(env, obj)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def DeleteWeakGlobalRef(self, ref):
        env = self.__env
        fun = env[0]
        fun.DeleteWeakGlobalRef(env, ref)

    def EnsureLocalCapacity(self, capacity):
        env = self.__env
        fun = env[0]
        ret = fun.EnsureLocalCapacity(env, capacity)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Java objects handling

    def AllocObject(self, clazz):
        env = self.__env
        fun = env[0]
        ret = fun.AllocObject(env, clazz)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewObject(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.NewObjectA(env, clazz, methodID, _itself_or_NULL(args))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetObjectClass(self, obj):
        env = self.__env
        fun = env[0]
        ret = fun.GetObjectClass(env, obj)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetObjectRefType(self, obj):
        # New in JNI 1.6
        env = self.__env
        fun = env[0]
        ret = fun.GetObjectRefType(env, obj)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def IsInstanceOf(self, obj, clazz):
        env = self.__env
        fun = env[0]
        return bool(fun.IsInstanceOf(env, obj, clazz))

    def IsSameObject(self, obj1, obj2):
        env = self.__env
        fun = env[0]
        return bool(fun.IsSameObject(env, obj1, obj2))

    # Call Java instance method

    def GetMethodID(self, clazz, name, sig):
        env = self.__env
        fun = env[0]
        ret = fun.GetMethodID(env, clazz, name, sig)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallObjectMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallObjectMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallBooleanMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallBooleanMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return bool(ret)

    def CallByteMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallByteMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallCharMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallCharMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallShortMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallShortMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallIntMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallIntMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallLongMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallLongMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallFloatMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallFloatMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallDoubleMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallDoubleMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallVoidMethod(self, obj, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        fun.CallVoidMethodA(env, obj, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # ... nonvirtually

    def CallNonvirtualObjectMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualObjectMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualBooleanMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualBooleanMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return bool(ret)

    def CallNonvirtualByteMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualByteMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualCharMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualCharMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualShortMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualShortMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualIntMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualIntMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualLongMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualLongMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualFloatMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualFloatMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualDoubleMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        ret = fun.CallNonvirtualDoubleMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallNonvirtualVoidMethod(self, obj, clazz, methodID, args=None):
        env = self.__env
        fun = env[0]
        fun.CallNonvirtualVoidMethodA(env, obj, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # Getting Java instance fields

    def GetFieldID(self, clazz, name, sig):
        env = self.__env
        fun = env[0]
        ret = fun.GetFieldID(env, clazz, name, sig)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetObjectField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetObjectField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetBooleanField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetBooleanField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return bool(ret)

    def GetByteField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetByteField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetCharField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetCharField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetShortField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetShortField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetIntField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetIntField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetLongField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetLongField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetFloatField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetFloatField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetDoubleField(self, obj, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetDoubleField(env, obj, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Setting Java instance fields

    def SetObjectField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetObjectField(env, obj, fieldID, _itself_or_NULL(value))
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetBooleanField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetBooleanField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetByteField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetByteField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetCharField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetCharField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetShortField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetShortField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetIntField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetIntField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetLongField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetLongField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetFloatField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetFloatField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetDoubleField(self, obj, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetDoubleField(env, obj, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # Call Java static method

    def GetStaticMethodID(self, clazz, name, sig):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticMethodID(env, clazz, name, sig)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticObjectMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticObjectMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticBooleanMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticBooleanMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return bool(ret)

    def CallStaticByteMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticByteMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticCharMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticCharMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticShortMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticShortMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticIntMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticIntMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticLongMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticLongMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticFloatMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticFloatMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticDoubleMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        ret = fun.CallStaticDoubleMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def CallStaticVoidMethod(self, clazz, methodID, args=None):  # (GIL ?)
        env = self.__env
        fun = env[0]
        fun.CallStaticVoidMethodA(env, clazz, methodID, _itself_or_NULL(args))
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # Getting Java static fields

    def GetStaticFieldID(self, clazz, name, sig):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticFieldID(env, clazz, name, sig)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticObjectField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticObjectField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticBooleanField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticBooleanField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return bool(ret)

    def GetStaticByteField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticByteField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticCharField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticCharField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticShortField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticShortField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticIntField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticIntField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticLongField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticLongField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticFloatField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticFloatField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStaticDoubleField(self, clazz, fieldID):
        env = self.__env
        fun = env[0]
        ret = fun.GetStaticDoubleField(env, clazz, fieldID)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Setting Java static fields

    def SetStaticObjectField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticObjectField(env, clazz, fieldID, _itself_or_NULL(value))
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticBooleanField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticBooleanField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticByteField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticByteField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticCharField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticCharField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticShortField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticShortField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticIntField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticIntField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticLongField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticLongField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticFloatField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticFloatField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetStaticDoubleField(self, clazz, fieldID, value):
        env = self.__env
        fun = env[0]
        fun.SetStaticDoubleField(env, clazz, fieldID, value)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # Java strings handling

    def NewString(self, unicode, slen):
        env = self.__env
        fun = env[0]
        ret = fun.NewString(env, unicode, slen)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStringLength(self, str):  # noqa: A002
        env = self.__env
        fun = env[0]
        ret = fun.GetStringLength(env, str)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStringChars(self, str, isCopy=None):  # noqa: A002
        env = self.__env
        fun = env[0]
        ret = fun.GetStringChars(env, str, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ReleaseStringChars(self, str, chars):  # noqa: A002
        env = self.__env
        fun = env[0]
        fun.ReleaseStringChars(env, str, chars)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def NewStringUTF(self, utf):
        env = self.__env
        fun = env[0]
        ret = fun.NewStringUTF(env, utf)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStringUTFLength(self, str):  # noqa: A002
        env = self.__env
        fun = env[0]
        ret = fun.GetStringUTFLength(env, str)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetStringUTFChars(self, str, isCopy=None):  # noqa: A002
        env = self.__env
        fun = env[0]
        ret = fun.GetStringUTFChars(env, str, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ReleaseStringUTFChars(self, str, chars):  # noqa: A002
        env = self.__env
        fun = env[0]
        fun.ReleaseStringUTFChars(env, str, chars)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetStringRegion(self, str, start, len, buf):  # noqa: A002
        env = self.__env
        fun = env[0]
        fun.GetStringRegion(env, str, start, len, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetStringUTFRegion(self, str, start, len, buf):  # noqa: A002
        env = self.__env
        fun = env[0]
        fun.GetStringUTFRegion(env, str, start, len, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # ... in a critical manner

    def GetStringCritical(self, string, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetStringCritical(env, string, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ReleaseStringCritical(self, string, cstring):
        env = self.__env
        fun = env[0]
        fun.ReleaseStringCritical(env, string, cstring)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # Java arrays handling

    def GetArrayLength(self, array):
        env = self.__env
        fun = env[0]
        ret = fun.GetArrayLength(env, array)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewObjectArray(self, size, clazz, init=None):
        env = self.__env
        fun = env[0]
        ret = fun.NewObjectArray(env, size, clazz, _itself_or_NULL(init))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetObjectArrayElement(self, array, index):
        env = self.__env
        fun = env[0]
        ret = fun.GetObjectArrayElement(env, array, index)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def SetObjectArrayElement(self, array, index, value):
        env = self.__env
        fun = env[0]
        fun.SetObjectArrayElement(env, array, index, _itself_or_NULL(value))
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def NewBooleanArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewBooleanArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewByteArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewByteArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewCharArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewCharArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewShortArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewShortArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewIntArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewIntArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewLongArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewLongArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewFloatArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewFloatArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def NewDoubleArray(self, size):
        env = self.__env
        fun = env[0]
        ret = fun.NewDoubleArray(env, size)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetBooleanArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetBooleanArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetByteArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetByteArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetCharArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetCharArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetShortArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetShortArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetIntArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetIntArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetLongArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetLongArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetFloatArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetFloatArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetDoubleArrayElements(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetDoubleArrayElements(env, array, _byref_or_NULL(isCopy))
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ReleaseBooleanArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseBooleanArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseByteArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseByteArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseCharArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseCharArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseShortArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseShortArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseIntArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseIntArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseLongArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseLongArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseFloatArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseFloatArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def ReleaseDoubleArrayElements(self, array, elems, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleaseDoubleArrayElements(env, array, elems, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetBooleanArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetBooleanArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetByteArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetByteArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetCharArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetCharArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetShortArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetShortArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetIntArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetIntArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetLongArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetLongArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetFloatArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetFloatArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def GetDoubleArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.GetDoubleArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetBooleanArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetBooleanArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetByteArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetByteArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetCharArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetCharArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetShortArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetShortArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetIntArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetIntArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetLongArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetLongArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetFloatArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetFloatArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    def SetDoubleArrayRegion(self, array, start, size, buf):
        env = self.__env
        fun = env[0]
        fun.SetDoubleArrayRegion(env, array, start, size, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # ... in a critical manner

    def GetPrimitiveArrayCritical(self, array, isCopy=None):
        env = self.__env
        fun = env[0]
        ret = fun.GetPrimitiveArrayCritical(env, array, _byref_or_NULL(isCopy))
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ReleasePrimitiveArrayCritical(self, array, carray, mode=0):
        env = self.__env
        fun = env[0]
        fun.ReleasePrimitiveArrayCritical(env, array, carray, mode)
        if fun.ExceptionCheck(env): self._handle_JavaException()

    # Java native methods handling

    def RegisterNatives(self, clazz, methods, nMethods):
        env = self.__env
        fun = env[0]
        # Required due to bug in jvm:
        # https://bugs.java.com/bugdatabase/view_bug.do?bug_id=6493522
        fun.GetMethodID(env, clazz, b"notify", b"()V")
        ret = 0  # fun.RegisterNatives(env, clazz, methods, nMethods)
        # print("RegisterNatives:ret:", ret)
        if ret != 0 and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def UnregisterNatives(self, clazz):
        env = self.__env
        fun = env[0]
        ret = fun.UnregisterNatives(env, clazz)
        if ret != 0 and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Java object monitoring

    def MonitorEnter(self, obj):
        env = self.__env
        fun = env[0]
        ret = fun.MonitorEnter(env, obj)
        if ret != 0 and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def MonitorExit(self, obj):
        env = self.__env
        fun = env[0]
        ret = fun.MonitorExit(env, obj)
        if ret != 0 and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Java direct buffer handling

    def NewDirectByteBuffer(self, address, capacity):
        env = self.__env
        fun = env[0]
        ret = fun.NewDirectByteBuffer(env, from_buffer(address), capacity)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetDirectBufferAddress(self, buf):
        env = self.__env
        fun = env[0]
        ret = fun.GetDirectBufferAddress(env, buf)
        if not ret and fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def GetDirectBufferCapacity(self, buf):
        env = self.__env
        fun = env[0]
        ret = fun.GetDirectBufferCapacity(env, buf)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Java reflection support

    def FromReflectedMethod(self, method):
        env = self.__env
        fun = env[0]
        ret = fun.FromReflectedMethod(env, method)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def FromReflectedField(self, field):
        env = self.__env
        fun = env[0]
        ret = fun.FromReflectedField(env, field)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ToReflectedMethod(self, cls, methodID, isStatic):
        env = self.__env
        fun = env[0]
        ret = fun.ToReflectedMethod(env, cls, methodID, isStatic)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    def ToReflectedField(self, cls, fieldID, isStatic):
        env = self.__env
        fun = env[0]
        ret = fun.ToReflectedField(env, cls, fieldID, isStatic)
        if fun.ExceptionCheck(env): self._handle_JavaException()
        return ret

    # Java VM Interface

    def GetJavaVM(self, vm):
        env = self.__env
        fun = env[0]
        ret = fun.GetJavaVM(env, vm)
        if ret != 0: self._handle_JNIException(ret)


tmap[JNIEnv] = ffi.typeof("JNIEnv")

def JEnv(penv):
    jenv = JNIEnv()
    jenv._JNIEnv__env = penv
    jenv._JNIEnv__fun = penv[0]
    return jenv

#
# JNI Invocation Interface.
#

class JavaVM:

    def _handle_JNIException(self, err):
        import sys
        fun_name = sys._getframe(1).f_code.co_name
        raise JNIException(err, info=fun_name)

    def DestroyJavaVM(self):
        jvm = self.__jvm
        fun = jvm[0]
        ret = fun.DestroyJavaVM(jvm)
        if ret != JNI_OK: self._handle_JNIException(ret)

    def AttachCurrentThread(self, penv, args=None):
        jvm = self.__jvm
        fun = jvm[0]
        p_penv = pointer(penv)
        ret = fun.AttachCurrentThread(jvm, cast(p_penv, "void**"), _itself_or_NULL(args))
        if ret != JNI_OK: self._handle_JNIException(ret)
        _as_CData(penv).c_data = _as_CData(p_penv[0]).c_data

    def AttachCurrentThreadAsDaemon(self, penv, args=None):
        jvm = self.__jvm
        fun = jvm[0]
        p_penv = pointer(penv)
        ret = fun.AttachCurrentThreadAsDaemon(jvm, cast(p_penv, "void**"), _itself_or_NULL(args))
        if ret != JNI_OK: self._handle_JNIException(ret)
        _as_CData(penv).c_data = _as_CData(p_penv[0]).c_data

    def DetachCurrentThread(self):
        jvm = self.__jvm
        fun = jvm[0]
        ret = fun.DetachCurrentThread(jvm)
        if ret != JNI_OK: self._handle_JNIException(ret)

    def GetEnv(self, penv, version):
        jvm = self.__jvm
        fun = jvm[0]
        p_penv = pointer(penv)
        ret = fun.GetEnv(jvm, cast(p_penv, "void**"), version)
        if ret != JNI_OK: self._handle_JNIException(ret)
        _as_CData(penv).c_data = _as_CData(p_penv[0]).c_data


tmap[JavaVM] = ffi.typeof("JavaVM")

def JVM(pjvm):
    jvm = JavaVM()
    jvm._JavaVM__jvm = pjvm
    jvm._JavaVM__fun = pjvm[0]
    return jvm


JavaVMOption = ffi.typeof("JavaVMOption")

JavaVMInitArgs = ffi.typeof("JavaVMInitArgs")

JavaVMAttachArgs = ffi.typeof("JavaVMAttachArgs")

# These will be VM-specific.

JDK1_2 = ffi.integer_const("JDK1_2")
JDK1_4 = ffi.integer_const("JDK1_4")

# End VM-specific.

# JNI version constants

JNI_VERSION_1_1 = ffi.integer_const("JNI_VERSION_1_1")
JNI_VERSION_1_2 = ffi.integer_const("JNI_VERSION_1_2")
JNI_VERSION_1_4 = ffi.integer_const("JNI_VERSION_1_4")
JNI_VERSION_1_6 = ffi.integer_const("JNI_VERSION_1_6")
JNI_VERSION_1_8 = ffi.integer_const("JNI_VERSION_1_8")
JNI_VERSION_9   = ffi.integer_const("JNI_VERSION_9")
JNI_VERSION_10  = ffi.integer_const("JNI_VERSION_10")
JNI_VERSION_19  = ffi.integer_const("JNI_VERSION_19")
JNI_VERSION_20  = ffi.integer_const("JNI_VERSION_20")
JNI_VERSION_21  = ffi.integer_const("JNI_VERSION_21")
JNI_VERSION_24  = ffi.integer_const("JNI_VERSION_24")

# eof jni.h

class Throwable(Exception):  # noqa: N818

    last = None

    def __init__(self, cause=NULL, info=NULL):  # noqa: D107
        self._cause = cast(cause, jthrowable)
        self._info  = cast(info,  jstring)
        super().__init__(self._cause, self._info)

    def getCause(self):
        return self._cause

    def getInfo(self):
        return self._info

class JNIException(SystemError):

    reason = {
        JNI_OK:        "success",
        JNI_ERR:       "unknown error",
        JNI_EDETACHED: "thread detached from the VM",
        JNI_EVERSION:  "JNI version error",
        JNI_ENOMEM:    "not enough memory",
        JNI_EEXIST:    "VM already created",
        JNI_EINVAL:    "invalid arguments",
    }

    def __init__(self, error=JNI_ERR, info=None):  # noqa: D107
        self._error = error
        self._info  = info
        super().__init__(self.getMessage(), self.getError())

    def getMessage(self):
        prefix = self._info + ": " if self._info else ""
        return prefix + JNIException.reason.get(self._error,
                                                f"unknown error code {self._error}")

    def getError(self):
        return self._error

def load(dll_path, handle=None, __dlclose=dlclose, __ffi=ffi):

    try:
        if isinstance(dll_path, os.PathLike): dll_path = str(dll_path)
        dll = DLL(dll_path, handle=handle)
    except OSError as exc:
        raise exc
    except Exception as exc:
        raise OSError(f"{exc}") from None

    def JNI_GetDefaultJavaVMInitArgs(args, __dll=dll):
        return __dll.JNI_GetDefaultJavaVMInitArgs(cast(args, "void*"))

    def JNI_CreateJavaVM(pvm, penv, args, __dll=dll):
        p_pvm  = pointer(pvm)
        p_penv = pointer(penv)
        p_args = pointer(args)
        result = __dll.JNI_CreateJavaVM(p_pvm, cast(p_penv, "void**"), cast(p_args, "void*"))
        _as_CData(pvm).c_data  = _as_CData(p_pvm[0]).c_data
        _as_CData(penv).c_data = _as_CData(p_penv[0]).c_data
        return result

    def JNI_GetCreatedJavaVMs(pvm, size, nvms, __dll=dll):
        return __dll.JNI_GetCreatedJavaVMs(_itself_or_NULL(pvm), size, nvms)

    JNI = type("JNI", (), dict(dll=dll, dllclose=classmethod(lambda cls, __dlclose=__dlclose: __dlclose(cls.dll))))
    JNI.GetDefaultJavaVMInitArgs = JNI_GetDefaultJavaVMInitArgs
    JNI.CreateJavaVM             = JNI_CreateJavaVM
    JNI.GetCreatedJavaVMs        = JNI_GetCreatedJavaVMs

    return JNI

#
# Decorator for defining Java native method in Python
#

def method(signature, **kwargs):
    ret_type, arg_types = __parse_signature(signature)
    fun_name = new_cstr(kwargs["name"].encode("utf-8")) if "name" in kwargs else None
    fun_sign = new_cstr(signature.encode("utf-8"))
    FunProto = CFUNC(ret_type, POINTER(JNIEnv), jobject, *arg_types)
    Fun_Proto = None
    _fun_name = None

    def wrapper(fun, fun_name=fun_name, fun_sign=fun_sign):
        nonlocal Fun_Proto, _fun_name
        Fun_Proto = FunProto(fun)
        _fun_name = new_cstr(fun.__name__.encode("utf-8"))
        native_meth = new(JNINativeMethod)[0]
        native_meth.name      = fun_name or _fun_name
        native_meth.signature = fun_sign
        native_meth.fnPtr     = cast(Fun_Proto, "void*")
        return native_meth

    return wrapper

def __parse_signature(signature):

    if signature[0] != "(":
        raise JNIException(JNI_EINVAL, info="jni.method")

    args_sig, sep, ret_sig = signature[1:].partition(")")
    if not sep or not ret_sig:
        raise JNIException(JNI_EINVAL, info="jni.method")

    sig = ret_sig + args_sig

    arg_types = []
    while sig:

        dim = 0
        while sig[dim] == "[": dim += 1
        sig = sig[dim:]

        ch = sig[0]
        if ch in "ZBCSIJFD" or (ch == "V" and dim == 0 and not arg_types):
            sig = sig[1:]
            if dim == 0:
                arg_types.append(__native_types[ch])
            elif dim == 1:
                arg_types.append(__native_array_types[ch])
            else:
                arg_types.append(jobjectArray)
        elif ch == "L":
            cname, sig = sig[1:].split(";", 1)
            if dim == 0:
                arg_types.append(__object_types.get(cname, jobject))
            else:
                arg_types.append(jobjectArray)
        else:
            raise JNIException(JNI_EINVAL, info="jni.method")

    return arg_types[0], tuple(arg_types[1:])


__native_types = {
    "V": None,
    "Z": jboolean,
    "B": jbyte,
    "C": jchar,
    "S": jshort,
    "I": jint,
    "J": jlong,
    "F": jfloat,
    "D": jdouble
}

__native_array_types = {
    "Z": jbooleanArray,
    "B": jbyteArray,
    "C": jcharArray,
    "S": jshortArray,
    "I": jintArray,
    "J": jlongArray,
    "F": jfloatArray,
    "D": jdoubleArray
}

__object_types = {
    "java/lang/String": jstring,
    "java/lang/Class":  jclass
}

del sys
del platform
del ct
del ffi
del tmap
del dlclose
